/*
    EepromHandler - Langlo EEPROM management
    Version 0.0.6	04/05/2023

    0.0.1 Initial release
    0.0.2 Change setAddressingMode() to initSmartSerial()
    0.0.3 Add getI2CAddress() method
          Expand EH_dataType to include MAC addresses and descriptor
          Add methods to get EH_dataType byte counts and locations
    0.0.4 Add dump()and scrub() methods
    0.0.5 Move softwareRevision() to Public
		0.0.6 Change const to static const/constexpr as required for LinkedList compatibility

    Digital Concepts
    May 2023
    digitalconcepts.net.au
   
    This is just a convenient way to keep all the information relating to EEPROM data storage
    in one place.
   
    The external EEPROM included on many of the PCBs is used, in spite of the fact that some
    processors include their own [versions of an] EEPROM, to store a sequence number, for
    packet sequencing, and counters currently used to record rainfall in the Weather
    Station application.
*/

#ifndef EepromHandler_h
#define EepromHandler_h

#include <Arduino.h>
#include <Wire.h>

// Index to EEPROM Data Types

enum EH_dataType {
  EH_GATEWAY_MAC = 0,  // 0
  EH_NODE_MAC,         // 1
  EH_DESCRIPTOR,       // 2
  EH_SEQUENCE,         // 3
  EH_RAINFALL,         // 4
  EH_NUM_TYPES
};

const uint8_t EH_MAX_BYTE_COUNT = 32;  // Set to largest value in _byteCount array

class EepromHandler {

  private:
  
/*
   The following two entities define the size and storage location respectively of the
   parameters stored in EEPROM
*/    
    static constexpr uint16_t _byteCount[EH_NUM_TYPES] = {
       4,       //  Gateway MAC
       4,       //  Node MAC
      32,       //  Descriptor
       2,       //  Sequence Number
       2,       //  Rainfall Counter
    };
    static const uint8_t _maxByteCount = EH_MAX_BYTE_COUNT;  // Largest value in _byteCount array
    static constexpr uint16_t _location[EH_NUM_TYPES] = {
        0,       //  Gateway MAC
        8,       //  Node MAC
       16,       //  Descriptor
      128,       //  Sequence Number
      132,       //  Rainfall Counter
    };

    static const uint8_t _defaultI2cAddress = 0x50;
    uint8_t _i2cAddress = _defaultI2cAddress;
    bool _EH_DEBUG = false;
    bool _smartSerial = true;
    TwoWire* _i2cBusPtr;

    union _eepromPayload {
      uint8_t payloadByte[_maxByteCount];
      uint32_t macAddress;
      uint16_t counter;
    };
    _eepromPayload _eepromData;

  public:

    EepromHandler();
    
    String softwareRevision();
    
    void init(TwoWire* _wirePtr, uint8_t _address);
    void begin(TwoWire* _wirePtr);
    void begin(TwoWire* _wirePtr, uint8_t _address);
    void begin(TwoWire* _wirePtr, bool _modeFlag);
    void begin(TwoWire* _wirePtr, uint8_t _address, bool _modeFlag);

    void setI2CBus(TwoWire* _wirePtr);
    void setI2CAddress(uint8_t _address);
    uint8_t getI2CAddress();
    bool initSmartSerial();
    void setSmartSerial(bool _value);
    bool getSmartSerial();
    void setDebug(bool _debug);
    bool getDebug();

    uint8_t getParameterByteCount(EH_dataType _dataType);
    uint8_t getParameterLocation(EH_dataType _dataType);
    
    void write(uint8_t* _buf, uint16_t _byteCount, uint16_t _eepromAddress);    
    void writeUint32(EH_dataType _dataType, uint32_t _value);    
    void writeUint16(EH_dataType _dataType, uint16_t _value);    
    void writeBytes(EH_dataType _dataType, uint8_t* _buf);    
    void writeBytes(EH_dataType _dataType, uint8_t* _buf, uint16_t _byteCount);
        
    void read(uint8_t* _buf, uint16_t _byteCount, uint16_t _eepromAddress);    
    uint32_t readUint32(EH_dataType _dataType);
    uint16_t readUint16(EH_dataType _dataType);
    uint8_t* readBytes(EH_dataType _dataType);
    
    void dump();
    void dump(uint16_t _startByte, uint16_t _endByte);
    void scrub();
    void scrub(uint16_t _startByte, uint16_t _endByte);
    
};

#endif
